/*------------------------------------------------------------------------------------------------------------
 * nano_scope
 * 
 * A more complex example showing how the GLCD works with Arduino.
 * This sketch implements a simplifed digital oszilloscope - just to show the principles.
 * A regular timer interrupt is running at 20KHz and monitors ADC values
 * A trigger logic detects raising edges of the signal and starts data collection. 
 * Once data collection is completed, the data is displayed as continous curve on the display.
 * The trigger logic is reset and ready for the next data collection.
 * A tiny keyboard (4 keys) allows to change time scale and trigger level 
 * 
 * Stephan Laage-Witt - 02-Jan-2018
 */

#include <Wire.h>
#include <glcd_functions.h>

// port pins
#define ADC_CHANNEL 0
#define LED_PIN 6
#define KEY1 2
#define KEY2 3
#define KEY3 4
#define KEY4 5
#define KEY_PORT_DDR DDRD
#define KEY_PORT PORTD

// constants
#define BUF_SIZE 158

// display positions
#define LABELS_X 175
#define Y_SCALE_LABEL_Y 3
#define TRG_LABEL_Y 22
#define TRG_Y 33
#define SAR_LABEL_Y 47
#define SAR_Y 56

// Global varibales ------------------------------------------------------------------------------------------
glcd gd(0x20);                                      // instance of the graphic display at I2C address 0x20
volatile uint8_t buf_a[BUF_SIZE], buf_b[BUF_SIZE];  // buffer for the adc values
uint8_t buf_pnt;                                    // buffer pointer 
uint8_t buf_select = 0;                             // buffer selection (0 -> buf_a, 1 -> buf_b)
volatile uint8_t trigger_flag = 0;                  // data coolection status, 0 -> waiting, 1 -> finished
volatile uint8_t key_pressed = 0;                   // keyboard status, any value > 0 -> a key was pressed) 
uint8_t trigger_voltage = 5;                        // trigger level in 0.1 V (0 ... 10)
uint8_t trigger_level;                              // converted trigger level (0 ... 255)

struct sample_rate_t {        
  uint16_t sar;                                     // sample rate
  uint8_t time_scale;                               // time per grid in msec
  uint8_t irq_cnt;                                  // number of interrupts per data sample item
};
const sample_rate_t sar_table[8] = {
  {200, 200, 100},
  {400, 100, 50},
  {800, 50, 25},
  {2000, 20, 10},
  {4000, 10, 5},
  {10000, 4, 2},
  {20000, 2, 1}
};
uint8_t sar_select = 3;

// setup function --------------------------------------------------------------------------------------------
void setup() {
  uint8_t x_cur, y_cur, font_height;
  
  // set ports
  pinMode(LED_PIN, OUTPUT);                               // output port for trigger indactor LED
  KEY_PORT_DDR &= ~((1 << KEY1) | (1 << KEY2) | (1 << KEY3) | (1 << KEY4));    // input port for buttons 
  KEY_PORT |= ((1 << KEY1) | (1 << KEY2) | (1 << KEY3) | (1 << KEY4));    // activate pull-up resistors
  
  // start I2C interface
  Wire.begin();
  
  // initiate display
  gd.set_light(10);
  gd.clear_screen();
  gd.set_font(5);
  gd.draw_str("Arduino Nano Scope");
  delay(30);
  gd.get_cursor(&x_cur, &y_cur);
  font_height = gd.get_font_height();
  gd.draw_line(0, font_height + 1, x_cur - 1, font_height + 1, 1);
  gd.set_font(4);
  gd.set_cursor(0, font_height + 5);
  gd.draw_str("Sample Rate: 20kHz");
  gd.set_font(0);
  
  // initiate ADC
  ADCSRA = (1<<ADEN) | (1<<ADPS2) | (0<<ADPS1) | (1<<ADPS0);    // scaling factor: 32 -> ADC input clock 500kHz 
  ADMUX  = (1<<REFS0) | (1<<REFS1) | (1<<ADLAR) | ADC_CHANNEL;  // left adjusted result, internal reference
  ADCSRA |= (1<<ADSC);                                    // start first ADC conversion
  
  // initiate timer1 for cyclic interrrupts at 20kHz
  TCNT1 = 0;
  TCCR1A = 0;
  TCCR1B = 0;
  TCCR1B |= (1 << WGM12);                                 // activate CTC mode
  TCCR1B |= (0 << CS12) | (1 << CS11) | (0 << CS10);      // set prescale to 8
  OCR1A = 100 - 1;    // set compare match register, interrupts at 16MHz / 8 / 100 -> 20kHz
  TIMSK1 |= (1 << OCIE1A);                                // enable timer inerrupts

  // draw static screen elements
  delay(1000);
  gd.clear_screen();
  gd.draw_line(BUF_SIZE, 0, BUF_SIZE, 63, 1);             // draw y-scale
  gd.set_font(1);
  for (int i = 0; i < 11; i += 2) {                       // draw tick marks
    int y_pos = 63 - (i * 64 / 11);
    gd.set_pixel(BUF_SIZE + 1, y_pos, 1);
    gd.set_pixel(BUF_SIZE + 2, y_pos, 1);
    if (i > 0) {
      gd.set_cursor(BUF_SIZE + 4, y_pos - 2);
      gd.draw_udec(i, 2, 1);  
    };
  };
  gd.set_cursor(LABELS_X, Y_SCALE_LABEL_Y);               // draw y-scale label
  gd.set_font(0);
  gd.draw_char('V');
  gd.set_font(4);
  gd.draw_line(LABELS_X, TRG_LABEL_Y - 3, 191, TRG_LABEL_Y - 3, 1); 
  gd.set_cursor(LABELS_X, TRG_LABEL_Y);
  gd.draw_str("Trig");                                    // print trigger voltage
  trigger_level = update_trigger_voltage(trigger_voltage);
  draw_grid(trigger_level);
  gd.draw_line(LABELS_X, SAR_LABEL_Y - 3, 191, SAR_LABEL_Y - 3, 1); 
  gd.set_cursor(LABELS_X, SAR_LABEL_Y);
  gd.draw_str("ms");                                      // print time scale
  print_time_scale(sar_table[sar_select].time_scale);

  // set system to "armed"
  sei();                  
}

// loop function ---------------------------------------------------------------------------------------------
void loop() {
  // process data collection by IRS
  if (trigger_flag > 0) {                             // IRS has finished data capturing, let's draw the curve
    if (buf_select == 0) {
      gd.draw_function(BUF_SIZE, 0, 0, buf_b);            // erase current data curve
      gd.draw_function(BUF_SIZE, 0, 1, buf_a);            // plot new curve
      buf_select = 1;                                     // swap data collection for the next round
    } else {
      gd.draw_function(BUF_SIZE, 0, 0, buf_a);            // erase current data curve
      gd.draw_function(BUF_SIZE, 0, 1, buf_b);            // plot new curve
      buf_select = 0;                                     // swap data collection for the next round
    };
    delay(100);                                           // wait a moment - screen refresh rate approximately 10Hz
    draw_grid(trigger_level);                             // refresh grid
    trigger_flag = 0;                                     // enable IRS to start next data collection
  };

  // process any input coming from the keyboard 
  if (key_pressed > 0) {                                  // system deteced user interaction
    switch (key_pressed) {                                // decode key pressed
      case 4:                                             // user pressed "sample rate down"
        if (sar_select > 0) {                             // decrease sample rate to next entry in table
          --sar_select;
          print_time_scale(sar_table[sar_select].time_scale);
        };
        break;
      case 8:                                             // user pressed "sample rate up"
        if (sar_select < (sizeof(sar_table) / sizeof(sample_rate_t)) - 2) {
          ++sar_select;                                   // increase sample rate to next entry in table
          print_time_scale(sar_table[sar_select].time_scale);
        };
        break;
      case 16:                                            // user pressed "trigger voltage down"
        if (trigger_voltage > 1) {
          --trigger_voltage;                              // decrease trigger voltage
          trigger_level = update_trigger_voltage(trigger_voltage);
        };
        break;
      case 32:                                            // user pressed "trigger voltage up"
        if (trigger_voltage < 10) {
          ++trigger_voltage;
          trigger_level = update_trigger_voltage(trigger_voltage);
        };
        break;
    };
    key_pressed = 0;
  };
}

/* print_time_scale ------------------------------------------------------------------------------------------
 * Sets the cursor and prints the current time per grid
 */
void print_time_scale(uint8_t time_scale) {
  gd.set_cursor(LABELS_X, SAR_Y);
  gd.draw_udec(time_scale, 3, 0);
  gd.draw_char(' ');
}

/* draw_grid -------------------------------------------------------------------------------------------------
 *  Plottes a grid of dotted lines. This includes refreshing the trigger line
 */
void draw_grid(uint8_t trg_level) {
  gd.draw_dotted_ver_line(40, 0, 63, 4, 1);
  gd.draw_dotted_ver_line(80, 0, 63, 4, 1);
  gd.draw_dotted_ver_line(120, 0, 63, 4, 1);
  draw_trigger_level(trg_level, 1);
}

/* update_trigger_voltage ------------------------------------------------------------------------------------
* Calculates the trigger value based on the trigger coltage.
* Refreshes the dotte line on the display indicating the triger level. 
* Updates the printed trigger value shown on the display
*/
uint8_t update_trigger_voltage(uint8_t trg_voltage) {
  static uint8_t old_trg_level = 0;
  uint8_t trg_level;

  draw_trigger_level(old_trg_level, 0);                   // clear current line
  trg_level = trg_voltage * 256 / 11;                     // calculate new trigger level
  draw_trigger_level(trg_level, 1);                       // draw new line
  gd.set_cursor(LABELS_X, TRG_Y);                         // show new value on display
  gd.draw_udec(trg_voltage, 2, 1);
  gd.draw_char(' ');
  old_trg_level = trg_level;
  return(trg_level);
}

/* draw_trigger_level ----------------------------------------------------------------------------------------
 * Draws a horizontal dotted line to indicate the trigger level
 */
void draw_trigger_level(uint8_t trg_level, uint8_t mode) {
  gd.draw_dotted_hor_line(63 - trg_level / 4, 0, BUF_SIZE - 1, 3, mode); // clear previous line
};

// ISR routine called by timer 1 interrupts ------------------------------------------------------------------
ISR(TIMER1_COMPA_vect){
  static uint8_t trigger_mode = 0;                        // trigger status (0 ... 3)
  static uint8_t adc_cnt = 0;                             // number of data points per data sample item
  static uint16_t adc_sum = 0;                            // sum of adc values per data sample item
  static uint8_t keyboard_mode = 0;                       // keyboard status (0 ... 10)
  static uint8_t prev_keyboard = 0;                       // keyboard value from previous round
  static uint16_t keyboard_cnt = 0;                       // keyboard is queried every nth cycle
  uint16_t adc_value, keyboard;

  // read and convert ADC value, restart ADC
  adc_sum += ADCH;                                        // add up adc values (upper 8 bits only, range 0 .. 255)
  ADCSRA |= (1 << ADSC);                                  // start next conversion

  if (++adc_cnt >= sar_table[sar_select].irq_cnt) {       // adc values are averaged across ...
    adc_value = adc_sum / adc_cnt;                        // ... capturing period
    adc_sum = 0;
    adc_cnt = 0; 

    // check trigger status ----------------------------------------------------------------------------------
    switch (trigger_mode) {
      case 0:                                             // waiting for trigger condition
        if (adc_value < trigger_level)                    // value needs to fall below threshold ...
          trigger_mode = 1;                               // ... to arm the trigger
          break;
      case 1:                                             // trigger is armed
        if (adc_value > trigger_level) {                  // value needs to raise above threshold ...
          if (buf_select == 0)                            // ... to enable data collection
            buf_a[0] = 63 - adc_value / 4;
          else
            buf_b[0] = 63 - adc_value / 4;
          buf_pnt = 1;
          trigger_mode = 2;
          digitalWrite(LED_PIN, HIGH);                    // indicate "trigger was activated"
        };
        break;
      case 2:                                             // data collection is running
        if (buf_select == 0)
          buf_a[buf_pnt] = 63 - adc_value / 4;
        else
          buf_b[buf_pnt] = 63 - adc_value / 4;
        if (++buf_pnt == BUF_SIZE) {
          trigger_flag = 1;
          trigger_mode = 3;
        };
        break;
      case 3:                                             // data collection completed ...
        if (trigger_flag == 0) {                          // ... waiting for main program to finish processing
          digitalWrite(LED_PIN, LOW);                     // clear trigger LED
          trigger_mode = 0;                               // reset trigger mode 
        };
        break;                                            
    }
  };
  
  // check keyboard ------------------------------------------------------------------------------------------
  if (++keyboard_cnt == 1000) {                           // this branch is executed 20 times per second
    keyboard_cnt = 0;
    keyboard = ~PIND & ((1 << KEY1) | (1 << KEY2) | (1 << KEY3) | (1 << KEY4));     // read keyboard
    if (keyboard > 0) {                                   // any key pressed?
      switch (keyboard_mode) {
        case 0:                                           // yes, a key stroke was recognized
          prev_keyboard = keyboard;
          ++keyboard_mode;
          break;
        case 2:                                           // still the same key? Then capture the key stroke
          if (prev_keyboard == keyboard) {
            key_pressed = keyboard;
            ++keyboard_mode;
          } else {
            keyboard_mode = 0;
          };
          break;
        case 10:                                          // same key pressed for at least 0.5 secs ... 
          if (prev_keyboard == keyboard) {                // ... triggers repeated-key feature
             key_pressed = keyboard;
             keyboard_mode = 5;
          } else {
            keyboard_mode = 0;   
          };
          break;
       default:                     
          if (prev_keyboard == keyboard) ++keyboard_mode;
          else keyboard_mode = 0;
      };
    } else {
      keyboard_mode = 0;
    };
  };
}

